import processing.opengl.*;
import javax.media.opengl.*;

long MILLIS_PER_DAY = 24 * 60 * 60 * 1000;

int WIDTH = 750;
int HEIGHT = 550;
int MARGIN = 20;
float ANGLE = 180.0;
int WINDOW = 58;
int OFFSET = 0;

int xAxisLength = 5000;
int yAxisLength = HEIGHT - 20;
int zAxisLength = 500;

float cameraX = 500; 
float cameraY = -200;
float cameraZ = 800;

float angleX = 0;
float angleY = 0; // 270
float angleZ = 0;

int cameraStep = 20; // Amount to move the camera with each key press.

PFont font24, font64;

ArrayList[][] data = new ArrayList[10][366]; // Every transaction. Format: [deweyCategory][day][transaction][x, y, z] (day includes potential leap year)
float[][][] data2 = new float[10][366][3]; // Just each day. Format: [deweyCategory][day][x, y, z]

Integrator currentDay;
Integrator axisColor;
Integrator menuColor;
Integrator calendarColor;
Integrator[] positions = new Integrator[3];

int currentGraph = 0;
int currentCategory = -1;
int currentTransaction = 0;
int amountTransactions = 20;
boolean showLines = false;
boolean showPoints = true;

boolean fadeCalendar = true;

void setup()
{
  size(WIDTH, HEIGHT, OPENGL);
  frame.setResizable(true);
  background(0);
  font24 = loadFont("GillSans-24.vlw");
  font64 = loadFont("GillSans-64.vlw");
  
  for(int i = 0; i < 10; i++)
  {
    for(int j = 0; j < 366; j++)
    {
      data[i][j] = new ArrayList();
    }
  }
  
  println("Checking for previously calculated data...");
  if(new File(dataPath("out.txt")).exists() && new File(dataPath("out2.txt")).exists())
  {
    println("Found. Loading data...");
    loadData2("out2.txt");
  }
  else
  {
    println("Not found. This will take a while.");
    println("Generating data...");
    processData("in.txt");
    println("Saving data...");
    saveData("out.txt");
    saveData2("out2.txt");
  }
  println("Done.");
  
  setupAnimations();
}

void setupAnimations()
{
  currentDay = new Integrator(0, 0.5, 0.001);
  currentDay.target(365);
  
  axisColor = new Integrator(192, 0.5, 0.002);
  axisColor.target(0);
  
  menuColor = new Integrator(128, 0.5, 0.1);
  menuColor.target(32);
  
  calendarColor = new Integrator(128, 0.5, 0.1);
  calendarColor.target(32);
  
  positions[0] = new Integrator(MARGIN);
  positions[1] = new Integrator(height - MARGIN);
  positions[2] = new Integrator(0);
}

void updateAnimations()
{
  currentDay.update();
  axisColor.update();
  if(millis() > 3000)
    menuColor.update();
  calendarColor.update();
}

void draw()
{
  background(0);
  updateAnimations();
  
  beginCamera();
  {
    pushMatrix();
    // Position camera at default position
    camera();
    
    // Move camera to initial position
    translate(-375.0, 225.0, -475.0);
    translate(cameraX, cameraY, cameraZ);
    
    // Rotate camera around x based on mouseY
    // Centers the mouse in the middle of the screen, and then gives a range from +ANGLE to -ANGLE
    angleX = ((mouseY - height/2.0) / ((height/2.0) / ANGLE)) / ANGLE * (ANGLE - abs(angleY));
    rotateX(radians(angleX));
    
    // Rotate camera around z based on mouseY
    angleZ = ((mouseY - height/2.0) / ((height/2.0) / ANGLE)) / ANGLE * -1*angleY;
    rotateZ(radians(angleZ));
    
    // Rotate camera around y based on mouseX
    angleY = -1 * (mouseX - width/2.0) / ((width/2.0) / ANGLE);
    rotateY(radians(angleY));
    
    popMatrix();
  }
  endCamera();

  drawGrid();
  drawData();
  updateCamera();

  camera(); // Reset camera.
  
  if(currentCategory != -1)
    drawTotals();
  drawMenu();
  
  fill(255);
  textFont(font24, 16);
  text("fps = " + frameRate, MARGIN, MARGIN + 5); //draw something in 2D
}

void drawGrid()
{
  smooth();
  
  pushMatrix();
    // Draw x
    stroke(axisColor.value);
    strokeWeight(1.2);
    translate(MARGIN, height - MARGIN, 0);
    line(0, 0, 0, xAxisLength, 0, 0);
  popMatrix();
  
  pushMatrix();
    // Draw y
    stroke(axisColor.value);
    strokeWeight(1.2);
    translate(MARGIN, height - MARGIN, 0);
    line(0, 0, 0, 0, -yAxisLength, 0);
  popMatrix();
  
  pushMatrix();
    // Draw z
    stroke(axisColor.value);
    strokeWeight(1.2);
    translate(MARGIN, height - MARGIN, 0);
    line(0, 0, 0, 0, 0, -zAxisLength);
  popMatrix();
  
  for(int i = 0; i < 10; i++) // For all ten sections...
  {
    pushMatrix();
      fill(getColor(i, 255));
      textFont(font24, 24);
      translate(MARGIN - 30, height - MARGIN, -(zAxisLength/10) * i);
      rotateY(radians(90));
      text(String.valueOf(i) + "00s", 0, 0, 0);
    popMatrix();
  }
  
  float xAxisStep = xAxisLength / 365.0;
  fill(128);
  pushMatrix();
    translate(MARGIN, height - MARGIN, 0);
    text("J", 0, 0, 0);
    text("F", xAxisStep * 32, 0, 0);
    text("M", xAxisStep * 61, 0, 0);
    text("A", xAxisStep * 92, 0, 0);
    text("M", xAxisStep * 122, 0, 0);
    text("J", xAxisStep * 153, 0, 0);
    text("J", xAxisStep * 183, 0, 0);
    text("A", xAxisStep * 214, 0, 0);
    text("S", xAxisStep * 245, 0, 0);
    text("O", xAxisStep * 275, 0, 0);
    text("N", xAxisStep * 306, 0, 0);
    text("D", xAxisStep * 336, 0, 0);
  popMatrix();
}

void drawData()
{
  if(currentGraph == 0)
    drawHollowLineGraph();
  else if(currentGraph == 1)
    drawTerrainGraph();
  else if(currentGraph == 2)
    drawPointGraph();
}

void drawHollowLineGraph()
{
  float yAxisStep = float(yAxisLength)/10000.0;
  
  for(int i = 0; i < data.length; i++)
  {
    stroke(getColor(i, 255));
    noFill();
    beginShape();
    
    for(int j = 0; j < currentDay.value; j++)
    {
      float[] c = data2[i][j];
      vertex(MARGIN + c[0], height - MARGIN - c[1], c[2]);
    }
    endShape();
    
    for(int j = 0; j < currentDay.value; j++)
    {
      float[] c = data2[i][j];
      fill(getColor(i, 200));
    
      pushMatrix();
        translate(MARGIN + c[0], height - MARGIN - c[1], c[2]);
        triangle(0, 0, -3, -3, 3, -3);
      popMatrix();
    }
  }
}

void drawTerrainGraph()
{
  float yAxisStep = float(yAxisLength)/10000.0;
  float zAxisStep = float(zAxisLength)/10.0;
  float halfzAxisStep = zAxisStep/2;
  
  for(int i = 0; i < data.length; i++)
  {
    for(float k = 0.0; k < zAxisStep; k += 1.0)
    {
      float percentComplete = 0;
      
      if(k < halfzAxisStep)
        percentComplete = k / halfzAxisStep;
      else
        percentComplete = 1 + ((halfzAxisStep - k) / halfzAxisStep);
      
      stroke( lerpColor(getColor(i, 50), getColor(i, 255), percentComplete) );
      fill( lerpColor(getColor(i, 50), getColor(i, 150), percentComplete) );
      
      beginShape();
      float[] c = data2[i][OFFSET];
      vertex(MARGIN + c[0], height - MARGIN, -(i*zAxisStep + k));
      for(int j = OFFSET; j < WINDOW + OFFSET; j++)
      {
        c = data2[i][j];
        vertex(MARGIN + c[0], height - MARGIN - lerp(0, c[1], percentComplete), -(i*zAxisStep + k));
      }
      c = data2[i][WINDOW + OFFSET - 1];
      vertex(MARGIN + c[0], height - MARGIN, -(i*zAxisStep + k));
      endShape();
    }
  }
}

void drawPointGraph()
{
  if(data[0][0].size() == 0)
  {
    loadData("out.txt");
  }
  
  float yAxisStep = float(yAxisLength)/10000.0;
  
  for(int i = 0; i < data.length; i++)
  {
    if(showLines)
    {
      stroke(getColor(i, 150));
      noFill();
      
      beginShape();
      for(int j = OFFSET; j < WINDOW + OFFSET; j++)
      {
        int dataLimit = amountTransactions;
        if(data[i][j].size() < dataLimit)
        {
          dataLimit = data[i][j].size();
        }
        for(int k = 0; k < dataLimit; k++)
        {
          float[] c = ((float[])data[i][j].get(k));
          curveVertex(MARGIN + c[0], height - MARGIN - c[1], c[2]);
        }
      }
      endShape();
    }
    
    if(showPoints)
    {
      for(int j = OFFSET; j < WINDOW + OFFSET; j++)
      {
        int dataLimit = amountTransactions;
        if(data[i][j].size() < dataLimit)
        {
          dataLimit = data[i][j].size();
        }
        for(int k = 0; k < dataLimit; k++)
        {
          float[] c = ((float[])data[i][j].get(k));
                  
          stroke(getColor(i, 150));
          fill(getColor(i, 150));
        
          pushMatrix();
            translate(MARGIN + c[0], height - MARGIN - c[1], c[2]);
            triangle(0, 0, -3, -3, 3, -3);
          popMatrix();
        }
      }
    }
  }
}

void updateCamera()
{
  if(currentCategory != -1)
  {
    // Handle if the graph is animating.
    if(currentGraph == 0 && currentDay.value < currentDay.target - 20)
      currentTransaction = int(currentDay.value);
    
    float[] c;
    if(currentGraph != 2)
    {
      c = data2[currentCategory][currentTransaction];
    }
    else
    {
      // Just in case there were no transactions for that specific day...
      // TODO: If you happen to be one day greater than the null day and try to go back, it keeps pushing you forward.
      while(data[currentCategory][currentTransaction].size() == 0)
        currentTransaction++;
      c = (float[])data[currentCategory][currentTransaction].get(0);
    }
    
    positions[0].target(c[0]);
    positions[1].target(c[1]);
    positions[2].target(c[2]);
    positions[0].update();
    positions[1].update();
    positions[2].update();
    
    cameraX = positions[0].value;
    cameraY = -positions[1].value;
    cameraZ = positions[2].value;
  }
}

void drawTotals()
{
  fill(getColor(currentCategory, 255));
  textFont(font24, 16);
  textAlign(RIGHT);
  text(currentCategory + "00s: " + getCategory(currentCategory), WIDTH - MARGIN, MARGIN + 5);
  text(int(data2[currentCategory][currentTransaction][1] / (yAxisLength/10000.0)), WIDTH - MARGIN, MARGIN + 25);
  textAlign(LEFT);
  
  drawCalendar();
}

void drawCalendar()
{
  int date[] = getDate(currentTransaction + 1);
  
  // Months are zero-indexed. Days are one-indexed. Java is stupid.
  Calendar calendar = new GregorianCalendar(2008, date[0], date[1]);
  calendar.setFirstDayOfWeek(Calendar.SUNDAY);
  int numberDays = calendar.getActualMaximum(GregorianCalendar.DAY_OF_MONTH);
  
  // Find the first day of the month. January 1st is a Tuesday, so offset by 2 (zero-indexed).
  int startDay = (currentTransaction - date[1] % 7 + 2 > 7) ? (currentTransaction - date[1] % 7 + 2) % 7 : currentTransaction - date[1] % 7 + 2;
  
  pushMatrix();
    fill(calendarColor.value);
    textAlign(RIGHT);
    text(getMonthName(date[0]), WIDTH - MARGIN, MARGIN * 4);
    text("*", WIDTH - 152, MARGIN * 4 + 5);
    translate(WIDTH - 152, MARGIN * 5);
    
    int currentRow = 0;
    for(int i = 0; i < numberDays; i++)
    {
      if(((i + startDay) % 7) == 0 && i != 0)
        currentRow++;
      
      if(isHoliday(date[0] + 1, i + 1))
        fill(255, 0, 0, calendarColor.value);
      else
        fill(calendarColor.value);
      text(i + 1, ((i + startDay) % 7)*22, currentRow*15);
      
      if(date[1] - 1 == i)
      {
        noFill();
        stroke(calendarColor.value);
        rect(((i + startDay) % 7)*22 - 18, currentRow*15 - 15, 20, 18);
      }
    }
    textAlign(LEFT);
  popMatrix();
  
  if(fadeCalendar && mouseX > WIDTH - 175 && mouseY > MARGIN * 4 - 15 && mouseY < MARGIN*4 + 85)
    calendarColor.target(128);
  else if(fadeCalendar)
    calendarColor.target(32);
  
}
  

void drawMenu()
{
  if(mouseY > HEIGHT - MARGIN - 30)
    menuColor.target(128);
  else
    menuColor.target(32);
  
  fill(menuColor.value);
  textFont(font24, 16);
  text("Line Graph", MARGIN + 5, HEIGHT - MARGIN - 5);
  if(currentGraph == 0 || (mouseX > MARGIN && mouseX < MARGIN + 82 && mouseY > HEIGHT - MARGIN*2 && mouseY < HEIGHT - MARGIN*2 + 20))
  {
    noFill();
    stroke(menuColor.value);
    rect(MARGIN, HEIGHT - MARGIN*2, 82, 20);
  }
  
  fill(menuColor.value);
  textFont(font24, 16);
  text("Terrain Graph", MARGIN + 95, HEIGHT - MARGIN - 5);
  if(currentGraph == 1 || (mouseX > MARGIN + 90 && mouseX < MARGIN + 195 && mouseY > HEIGHT - MARGIN*2 && mouseY < HEIGHT - MARGIN*2 + 20))
  {
    noFill();
    stroke(menuColor.value);
    rect(MARGIN + 90, HEIGHT - MARGIN*2, 105, 20);
  }
  
  fill(menuColor.value);
  textFont(font24, 16);
  text("Point Graph", MARGIN + 205, HEIGHT - MARGIN - 5);
  if(currentGraph == 2 || (mouseX > MARGIN + 200 && mouseX < MARGIN + 290 && mouseY > HEIGHT - MARGIN*2 && mouseY < HEIGHT - MARGIN*2 + 20))
  {
    noFill();
    stroke(menuColor.value);
    rect(MARGIN + 200, HEIGHT - MARGIN*2, 90, 20);
  }
  if(currentGraph == 2)
  {
    if(mouseY > HEIGHT - MARGIN*2 - 50)
      menuColor.target(128);
    else
      menuColor.target(32);
    
    fill(menuColor.value);
    textFont(font24, 16);
    text("Show Lines", MARGIN + 205, HEIGHT - MARGIN - 5 - 25);
    if(showLines == true || (mouseX > MARGIN + 200 && mouseX < MARGIN + 290 && mouseY > HEIGHT - MARGIN*2 - 25 && mouseY < HEIGHT - MARGIN*2))
    {
      noFill();
      stroke(menuColor.value);
      rect(MARGIN + 200, HEIGHT - MARGIN*2 - 25, 90, 20);
    }
    
    fill(menuColor.value);
    textFont(font24, 16);
    text("Show Points", MARGIN + 205, HEIGHT - MARGIN - 5 - 50);
    if(showPoints == true || (mouseX > MARGIN + 200 && mouseX < MARGIN + 290 && mouseY > HEIGHT - MARGIN*2 - 50 && mouseY < HEIGHT - MARGIN*2 - 25))
    {
      noFill();
      stroke(menuColor.value);
      rect(MARGIN + 200, HEIGHT - MARGIN*2 - 50, 90, 20);
    }
  }
  
  if(currentGraph == 0)
  {
    fill(menuColor.value);
    textFont(font24, 16);
    text("Animate", WIDTH - MARGIN - 60, HEIGHT - MARGIN - 5);
    if(mouseX > WIDTH - MARGIN - 65 && mouseX < WIDTH - MARGIN && mouseY > HEIGHT - MARGIN*2 && mouseY < HEIGHT - MARGIN*2 + 20)
    {
      noFill();
      stroke(menuColor.value);
      rect(WIDTH - MARGIN - 65, HEIGHT - MARGIN*2, 65, 20);
    }
  }
}

void keyPressed()
{
  if(key == 'a')
  {
    cameraZ += (cameraStep / ANGLE) * (angleY);
    cameraX -= (cameraStep / ANGLE) * (ANGLE - abs(angleY));
    currentCategory = -1;
    currentTransaction = 0;
  }
  else if(key == 'd')
  {
    cameraZ -= (cameraStep / ANGLE) * (angleY);
    cameraX += (cameraStep / ANGLE) * (ANGLE - abs(angleY));
    currentCategory = -1;
    currentTransaction = 0;
  }
  else if(key == 's')
  {
    cameraZ += (cameraStep / ANGLE) * (ANGLE - abs(angleY));
    cameraX -= (cameraStep / ANGLE) * (-1 * angleY);
    currentCategory = -1;
    currentTransaction = 0;
  }
  else if(key == 'w')
  {
    cameraZ -= (cameraStep / ANGLE) * (ANGLE - abs(angleY));
    cameraX += (cameraStep / ANGLE) * (-1 * angleY);
    currentCategory = -1;
    currentTransaction = 0;
  }
  else if(key == 'q')
  {
    cameraY += cameraStep;
    currentCategory = -1;
    currentTransaction = 0;
  }
  else if(key == 'e')
  {
    cameraY -= cameraStep;
    currentCategory = -1;
    currentTransaction = 0;
  }
  
  else if(key == ',' || key == '<')
  {
    if(currentTransaction > 0)
      currentTransaction--;
  }
  else if(key == '.' || key == '>')
  {
    if(data2[currentCategory].length > currentTransaction)
      currentTransaction++;
  }
  else if(key == '-')
  {
    if(amountTransactions > 0)
      amountTransactions--;
  }
  else if(key == '=' || key == '+')
  {
    amountTransactions++;
  }
  else if(key == '0' || key == '1' || key == '2' || key == '3' || key == '4' || key == '5' || key == '6' || key == '7' || key == '8' || key == '9')
  {
    currentCategory = Character.getNumericValue(key);
    positions[0].set(cameraX);
    positions[1].set(-cameraY);
    positions[2].set(cameraZ);
  }
      
  if(key == CODED)
  { 
    if(keyCode == UP && (WINDOW + OFFSET) < 365)
    {
      WINDOW++;
    }
    else if(keyCode == DOWN && WINDOW > 0)
    {
      WINDOW--;
    }
    else if(keyCode == LEFT && OFFSET > 0)
    {
      OFFSET--;
    }
    else if(keyCode == RIGHT && (WINDOW + OFFSET) < 365)
    {
      OFFSET++;
    }
  }
}

void mousePressed()
{
  if(mouseX > MARGIN && mouseX < MARGIN + 82 && mouseY > HEIGHT - MARGIN*2 && mouseY < HEIGHT - MARGIN*2 + 20)
  {
    currentGraph = 0;
  }
  else if(mouseX > MARGIN + 90 && mouseX < MARGIN + 195 && mouseY > HEIGHT - MARGIN*2 && mouseY < HEIGHT - MARGIN*2 + 20)
  {
    currentGraph = 1;
  }
  else if(mouseX > MARGIN + 200 && mouseX < MARGIN + 290 && mouseY > HEIGHT - MARGIN*2 && mouseY < HEIGHT - MARGIN*2 + 20)
  {
    currentGraph = 2;
    
    if(data[0][0].size() == 0)
    {
      fill(255);
      stroke(255);
      textFont(font24, 24);
      textAlign(CENTER);
      text("Loading additional data...", WIDTH / 2, HEIGHT / 2);
      textAlign(LEFT);
    }
  }
  else if(currentGraph == 2)
  {
    if(mouseX > MARGIN + 200 && mouseX < MARGIN + 290 && mouseY > HEIGHT - MARGIN*2 - 25 && mouseY < HEIGHT - MARGIN*2)
      showLines = !showLines;
    if(mouseX > MARGIN + 200 && mouseX < MARGIN + 290 && mouseY > HEIGHT - MARGIN*2 - 50 && mouseY < HEIGHT - MARGIN*2 - 25)
      showPoints = !showPoints;
  }  
  else if(mouseX > WIDTH - MARGIN - 65 && mouseX < WIDTH - MARGIN && mouseY > HEIGHT - MARGIN*2 && mouseY < HEIGHT - MARGIN*2 + 20)
  {
    currentDay.value = 0;
  }
  else if(mouseX > WIDTH - 162 && mouseX < WIDTH - 146 && mouseY > MARGIN * 4 - 16 && mouseY < MARGIN * 4)
  {
    fadeCalendar = !fadeCalendar;
  }
}

color getColor(int category, int alpha)
{
  switch(category)
  {
    case 0:
      return color(0, 110, 186, alpha);
    case 1:
      return color(141, 28, 12, alpha);
    case 2:
      return color(62, 166, 59, alpha);
    case 3:
      return color(184, 0, 138, alpha);
    case 4:
      return color(128, 128, 128, alpha);
    case 5:
      return color(255, 232, 18, alpha);
    case 6:
      return color(255, 109, 6, alpha);
    case 7:
      return color(223, 27, 21, alpha);
    case 8:
      return color(43, 96, 136, alpha);
    case 9:
      return color(138, 0, 184, alpha);
    default:
      return color(255, alpha);
  }
}

String getCategory(int category)
{
  switch(category)
  {
    case 0:
      return "Generalities";
    case 1:
      return "Philosophy and Psychology";
    case 2:
      return "Religion";
    case 3:
      return "Social Sciences";
    case 4:
      return "Language";
    case 5:
      return "Natural Sciences & Mathematics";
    case 6:
      return "Technology";
    case 7:
      return "The Arts";
    case 8:
      return "Literature & Rhetoric";
    case 9:
      return "Geography & History";
    default:
      return "";
  }
}

int[] getDate(int elapsedDays)
{
  if(elapsedDays < 32)
    return new int[] {0, elapsedDays};
  else if(elapsedDays < 61)
    return new int[] {1, elapsedDays - 31};
  else if(elapsedDays < 92)
    return new int[] {2, elapsedDays - 60};
  else if(elapsedDays < 122)
    return new int[] {3, elapsedDays - 91};
  else if(elapsedDays < 153)
    return new int[] {4, elapsedDays - 121};
  else if(elapsedDays < 183)
    return new int[] {5, elapsedDays - 152};
  else if(elapsedDays < 214)
    return new int[] {6, elapsedDays - 182};
  else if(elapsedDays < 245)
    return new int[] {7, elapsedDays - 213};
  else if(elapsedDays < 275)
    return new int[] {8, elapsedDays - 244};
  else if(elapsedDays < 306)
    return new int[] {9, elapsedDays - 274};
  else if(elapsedDays < 336)
    return new int[] {10, elapsedDays - 305};
  else
    return new int[] {11, elapsedDays - 335};
}

String getMonthName(int monthValue)
{
  switch(monthValue)
  {
    case 0:
      return "January";
    case 1:
      return "February";
    case 2:
      return "March";
    case 3:
      return "April";
    case 4:
      return "May";
    case 5:
      return "June";
    case 6:
      return "July";
    case 7:
      return "August";
    case 8:
      return "September";
    case 9:
      return "October";
    case 10:
      return "November";
    case 11:
      return "December";
    default:
      return "";
  }
}

boolean isHoliday(int monthValue, int dayValue)
{
  if(monthValue == 1 && dayValue == 1)
    return true;
  else if(monthValue == 1 && dayValue == 21)
    return true;
  else if(monthValue == 2 && dayValue == 14)
    return true;
  else if(monthValue == 2 && dayValue == 18)
    return true;
  else if(monthValue == 3 && dayValue == 23)
    return true;
  else if(monthValue == 5 && dayValue == 26)
    return true;
  else if(monthValue == 7 && dayValue == 4)
    return true;
  else if(monthValue == 9 && dayValue == 1)
    return true;
  else if(monthValue == 10 && dayValue == 13)
    return true;
  else if(monthValue == 10 && dayValue == 31)
    return true;
  else if(monthValue == 11 && dayValue == 11)
    return true;
  else if(monthValue == 11 && dayValue == 27)
    return true;
  else if(monthValue == 12 && dayValue == 24)
    return true;
  else if(monthValue == 12 && dayValue == 25)
    return true;
  else
    return false;
}

void processData(String fileName)
{
  String[] lines = loadStrings(fileName);
  int[][] dayCount = new int[10][366]; // Damn leap years...
  
  println("Calculating totals...");
  for(int i = 0; i < lines.length; i++)
  { 
    Transaction t = parseTransaction(lines[i]);
    dayCount[t.majorCategory][getElapsedDays(t.month, t.day)]++;
  }
  
  println("Calculating axis placement...");
  for(int i = 0; i < lines.length; i++)
  {
    Transaction t = parseTransaction(lines[i]);
    int elapsedDays = getElapsedDays(t.month, t.day);
    
    float x = (xAxisLength/365.0) * ((float)elapsedDays + float(t.hour)/24 + float(t.minute)/1440.0);
    float y = (yAxisLength/10000.0) * float(dayCount[t.majorCategory][elapsedDays]);
    float z = -(zAxisLength/10.0) * t.majorCategory - ((float)t.minorCategory/99.0) * (zAxisLength/10);
    
    data[t.majorCategory][elapsedDays].add(new float[] {x, y, z});
    
    lines[i] = null; // Tell the garbage collector to free the memory ASAP;
    t = null;
  }
  
  for(int i = 0; i < 10; i++)
  {
    for(int j = 0; j < 366; j++)
    {
      float x = (xAxisLength/365.0) * j;
      float y = (yAxisLength/10000.0) * float(dayCount[i][j]);
      float z = -(zAxisLength/10.0) * i + -(zAxisLength/20.0);
      
      data2[i][j][0] = x;
      data2[i][j][1] = y;
      data2[i][j][2] = z;
    }
  }
}

void loadData(String fileName)
{
  String[] lines = loadStrings(fileName);
  
  for(int i = 0; i < lines.length; i++)
  {
    float[] sections = float(split(lines[i], ","));
    data[int(sections[0])][int(sections[1])].add(new float[] {sections[2], sections[3], sections[4]});
  }
}

void loadData2(String fileName)
{
  String[] lines = loadStrings(fileName);
  
  for(int i = 0; i < lines.length; i++)
  {
    float[] sections = float(split(lines[i], ","));
    data2[int(sections[0])][int(sections[1])][0] = sections[2];
    data2[int(sections[0])][int(sections[1])][1] = sections[3];
    data2[int(sections[0])][int(sections[1])][2] = sections[4];
  }
}

void saveData(String fileName)
{
  PrintWriter out = createWriter(dataPath(fileName));
  String datum = "";
  
  for(int i = 0; i < data.length; i++)
  {
    for(int j = 0; j < data[i].length; j++)
    {
      for(int k = 0; k < data[i][j].size(); k++)
      {
        datum = String.valueOf(i) + "," + String.valueOf(j) + "," + String.valueOf( ((float[])data[i][j].get(k))[0] ) + "," + String.valueOf( ((float[])data[i][j].get(k))[1] ) + "," + String.valueOf( ((float[])data[i][j].get(k))[2] );
        out.println(datum);
      }
    }
  }
  
  out.flush();
  out.close();
}

void saveData2(String fileName)
{
  PrintWriter out = createWriter(dataPath(fileName));
  String datum = "";
  
  for(int i = 0; i < data2.length; i++)
  {
    for(int j = 0; j < data2[i].length; j++)
    {
      datum = String.valueOf(i) + "," + String.valueOf(j) + "," + String.valueOf(data2[i][j][0]) + "," + String.valueOf(data2[i][j][1]) + "," + String.valueOf(data2[i][j][2]);
      out.println(datum);
    }
  }
  
  out.flush();
  out.close();
}

int getElapsedDays(int m, int d)
{
  Date firstDate = new GregorianCalendar(2008, 1, 1).getTime(); // getTime() should actually be getDate(), but Java is stupid.
  Date secondDate = new GregorianCalendar(2008, m, d).getTime();
  
  return new Long((secondDate.getTime() - firstDate.getTime()) / MILLIS_PER_DAY).intValue();
}

public Transaction parseTransaction(String line)
{
  String[] sections = split(line, ",");

  Transaction t = new Transaction();
  
  String[] date = split(sections[0], "-");
  String[] time = split(sections[1], ":");
  String majorCategory = sections[2].substring(0, 1);
  String minorCategory = sections[2].substring(1, 3);
  
  t.majorCategory = Integer.parseInt(majorCategory);
  t.minorCategory = Integer.parseInt(minorCategory);
  t.month = Integer.parseInt(date[1]);
  t.day = Integer.parseInt(date[2]);
  t.hour = Integer.parseInt(time[0]);
  t.minute = Integer.parseInt(time[1]);
  
  return t;
}
